//
// Implementation of ControlWindow class
//
// this belongs to the 'Tower of Benoi' Project
// Copyright 1998-1999
//
// Written by: Marco Zinn
//

#include <stdio.h>
#include <string.h>
#include "Towers.h" // we need some of the global stuff
#include "ControlWindow.h"
#include "TowerWindow.h" // we need to know, how the TowerWindow is defined

// The constructor sets up the Window
//  This includes the Slides and the Button
ControlWindow::ControlWindow(BRect frame, const char* name, DiscTowerView *FromTower, DiscTowerView *ToTower, DiscTowerView *BufferTower)
			: BWindow(frame, name,B_TITLED_WINDOW,B_NOT_ZOOMABLE|B_NOT_RESIZABLE|B_WILL_ACCEPT_FIRST_CLICK|B_NOT_CLOSABLE) {

	From   = FromTower;  // Save the Pointers to the three Towers for later
	To     = ToTower;
	Buffer = BufferTower;

	// Create Slider. Note that r.bottom is set equal to r.top
	//  We can do this, because it's ignored for BSliders.
	BRect r(10,20,Bounds().Width()-10,20);
	
	// Create View with Hashmarks and labels
	SpeedSlider=new BSlider(r,"speed","Solving Speed",new BMessage(),0,100);
	SpeedSlider->SetHashMarks(B_HASH_MARKS_BOTTOM);
	SpeedSlider->SetLimitLabels("Slow", "Fast");
	SpeedSlider->SetValue(50); // Set it to the middle position
	AddChild(SpeedSlider); // and add it to the window
	
	r.OffsetBy(0,60); // go some pixels lower
	ToggleButton = new BButton(r,"start","Start Auto-Solve",new BMessage(START_SOLVING));
	ToggleButton->SetTarget(be_app);
	AddChild(ToggleButton);
		// Create and add button including the model message

	StopFlag=false; // Don't stop the solving (hey, we didn't start yet!)
	SolvingSem = create_sem(1, "solver_sem"); // Create Semaphore for solver thread
	Show();
	// Note that the window is not shown by default!
}


// BuildTower
status_t ControlWindow::BuildTower(DiscTowerView *ToTower, int NumDiscs) {
	printf("Entering BuildTower: I shall build %i Discs on %c\n",NumDiscs,ToTower->GetName()); // Debug

	printf("-Calling MoveDiscTo with %i \n",NumDiscs); // Debug
	MoveDiscTo(ToTower,NumDiscs); // Move biggest disc to target Tower

	// Check if the solving must be stopped.
	//  Stop, when the StopFlag is on (true).
	if (StopFlag) { printf("!Stop requested in Buildtower, so leaving before getting too deep.\n"); return(B_NO_ERROR);}


	if (NumDiscs>1) { // Stack the other discs on top, if any.
		printf("-Calling MoveTower with %i \n",NumDiscs-1); // Debug
		BuildTower(ToTower,NumDiscs-1); // Stack the rest of the discs on top of the big one
		}

	printf("Leaving BuildTower: Tower with %i Discs has been built on %c\n",NumDiscs,ToTower->GetName()); // Debug
	return(B_NO_ERROR);
	}

status_t ControlWindow::MoveDiscTo(DiscTowerView *ToTower, int DiscNr) {
	printf("Entering MoveDiscTo: I shall move Disc %i to %c\n",DiscNr,ToTower->GetName()); // Debug

	// Check if the solving must be stopped.
	//  Stop, when the StopFlag is on (true).
	if (StopFlag) { printf("!Stop requested in MoveDisc, so leaving before really starting.\n"); return(B_NO_ERROR);}

		
	if (ToTower->HasItem(Discs[DiscNr])) { // Oh, we're lucky....
		printf("Great, the tower already has it.\n"); // Debug
		return(B_NO_ERROR);
		}
	
	// Well, first we need to find out, where the Discs we are searching for is right now.
	DiscTowerView *FromTower=NULL;
	TowerWindow *TW=((TowerApp*)be_app)->Towers;
	
	if (TW->LeftTowerView->HasItem(Discs[DiscNr])) 
		FromTower=TW->LeftTowerView;
	if (TW->CenterTowerView->HasItem(Discs[DiscNr]))
		FromTower=TW->CenterTowerView;
	if (TW->RightTowerView->HasItem(Discs[DiscNr]))
		FromTower=TW->RightTowerView;
	
	if (FromTower==NULL) { // ERROR: The disc was not on any Tower (it's not in play!)
		printf("Shit, couldn't locate the disc %i.\n",DiscNr); // Debug
		return(B_ERROR);
		}

	// Fine, we now know, where the disc is. In Order to to move the disc from 'FromTower' to 'ToTower',
	// we need make space on top of both towers. 'FromTower' must be empty above the disc, that we want to move.
	// 'ToTower' must not contain any smaller discs than the one we are moving, as this would block that tower 
	// (remember the rules: no bigger disc may be on a smaller one!)
	// As this implies that all discs, that are smaller than our to-move disc, must be 'out of the way', they must be placed at
	// the remaining tower. They must be in order (remember the rule, again). This means, that we have to build a small
	// stack of discs  at the remaining tower place.
	// Example: Moving Disc 5 from A to C means, that Disc 5 must be the uppermost disc on Tower A, that Tower C must not
	// contain any of the smaller discs (1 through 4) --> All smaller discs (1-4) must be stacked at Tower B. Got it?

	if (DiscNr>1) { // Check, if we need to create a buffer tower at all.
					// If we're moving Disc 1, there's no need to do build a buffer tower at all.
		DiscTowerView *BufferTower=NULL; // Pointer to the buffer Tower
	
		if (FromTower==TW->LeftTowerView) { // Find Buffer Tower, if disc is on the left Tower
			if (ToTower==TW->CenterTowerView)
				BufferTower=TW->RightTowerView;
			else
				BufferTower=TW->CenterTowerView;
			}
	
		if (FromTower==TW->CenterTowerView) { // Find Buffer Tower, if disc is on the center Tower
			if (ToTower==TW->LeftTowerView)
				BufferTower=TW->RightTowerView;
			else
				BufferTower=TW->LeftTowerView;
			}
	
		if (FromTower==TW->RightTowerView) { // Find Buffer Tower, if disc is on the right Tower
			if (ToTower==TW->LeftTowerView)
				BufferTower=TW->CenterTowerView;
			else
				BufferTower=TW->LeftTowerView;
			}
		
		if (FromTower==NULL) { // ERROR: Couldn't find a Buffer Tower
			printf("!Couldn't find a Buffer Tower.\n"); // Debug
			return(B_ERROR);
			}
	
		BuildTower(BufferTower,DiscNr-1); // Create the Tower of 'smaller discs' on the Buffer Tower
	}

	// Check if the solving must be stopped.
	//  Stop, when the StopFlag is on (true).
	if (StopFlag) { printf("!Stop requested in MoveDisc, so leaving.\n"); return(B_NO_ERROR);}

	// Now, move the disc
	printf("-Moving the Disc No. %i (On Tower: %i)\n",DiscNr,FromTower->FirstDiscPtr()->size); // Debug
	BMessage messy(OFFER_DISC);
		// Moving the Disc is done via the messaging mechanism and protocol, that is
		// used during the drag-and-drop-operations. So generate a message.
	messy.AddPointer("DiscPtr",FromTower->FirstDiscPtr());
		// Add a Pointer to the uppermost Disc of the 'source' Tower to the message
	ToTower->Window()->PostMessage(&messy,(BHandler*)ToTower,(BHandler*)FromTower);
		// That's the clue:
		// This posts a message to 'ToTower' (the Destination Tower for the Disc)
		// The message offers the uppermost disc of the 'FromTower'
		// the 2nd argument '(BHandler*)ToTower' specifies the ToTower BHandler
		//   (that is the (DiscTower-)View) as the target handler (ie, that view gets the message)
		// the 3rd argument '(BHandler*)FromTower' specifies the FromTower BHandler
		//   (again the (DiscTower-)View) as the reply handler (ie, that view gets any replies to the message)
		// This means, the ToTower 'thinks', the 'FromTower' did send the message and replies to it.
		// That's great, as we are out of the whole stuff and can leave the Towers do the rest.
	
	// printf("-Delaying\n");  // Debug
	snooze((103-SpeedSlider->Value())*15000); // Delay (value in microseconds)
	// This sets our thread (the ControlWindow thread) to sleep for some time
	// NOTE: that there is some delay necessary for the messages to be acted upon.
	//       If there's no delay, the whole thing will lock up!
	//       Probably, that's because the messages are posted to fast, that they cannot
	//       be replied to in time and the towers get inconsistent. 
	//       You will see one disc on two towers, when this happens!

	printf("Leaving MoveDiscTo:  Disc %i is now on %c\n",DiscNr,ToTower->GetName()); // Debug
	
	return B_NO_ERROR;
}



// SolveGame solves the game
//  It relies on somebody (the application in our case) to have reset the towers,
//  as it depends on that, but it can't reset them itself.
//  While the solving is beeing done, it shows the window
//  After finishing the solving (or, when it's stopped somehow), it hides the window again
status_t ControlWindow::SolveGame() {
	StopFlag=false; // well, nice idea, isn't it.
	SolvingThread = spawn_thread(thread_func, "Solver", B_LOW_PRIORITY, this); 
	resume_thread(SolvingThread); 
	return B_NO_ERROR;
}

int32 ControlWindow::thread_func(void *arg) {  // static function!
	printf("Trying to start the solver thread...\n"); // Debug
	if (acquire_sem(SolvingSem) != B_NO_ERROR) { return B_ERROR; }
	ControlWindow *obj = (ControlWindow *)arg;
	printf("* Starting Thread: Auto Solving\n"); // Debug
	obj->BuildTower(obj->To,NumberOfDiscs);
	printf("* EXITING Thread: Auto Solving\n"); // Debug
	release_sem(SolvingSem);
	return (B_NO_ERROR); 
}

// StopSolving simply sets the Flag true.
//  When this flag is true, the recursive function (MoveTower) will returm immediately
//  and thus the solving will stop nearly at once.
void ControlWindow::StopSolving() {
	StopFlag=true;
	if (acquire_sem(SolvingSem) != B_NO_ERROR) return; // Wait for the solver thread to exit
	release_sem(SolvingSem);
}

void ControlWindow::Reset() {
	StopSolving();
	Lock();
	ToggleButton->SetLabel("Start Auto-Solve");
	ToggleButton->SetMessage(new BMessage(START_SOLVING));
	Unlock();
}	

void ControlWindow::StartSolving() {
	printf("StartSolving called...\n");
	Lock();
	ToggleButton->SetLabel("Stop Auto-Solve");
	ToggleButton->SetMessage(new BMessage(STOP_SOLVING));
	Unlock();
	SolveGame();
}

