//
// Implementation of HistoryWindow Classes
//
// this belongs to the 'Tower of Benoi' Project
// Copyright 1998-1999
//
// Written by: Marco Zinn
//

#include "HistoryWindow.h"
#include "Towers.h"
#include <stdio.h>
#include <string.h>
//#include <Application.h>

class HistoryLine : public BStringItem {
	public:
		HistoryLine(int32 InitNr, int InitFromTower, int InitToTower, int InitDiscSize, int32 InitCheckedAgainst);
		~HistoryLine() { printf("HistoryLine destroyed!\n"); }
		void MarkAsUndoMove();
		bool IsUndoMove(HistoryLine *AnotherLine);
		int32 CheckedAgainst; // Nr. of Move against which this hase been checked (=LastEffectiveMove, when this move was done)
	private:
		int32 Nr;
		int FromTower;
		int ToTower;
		int DiscSize;
};

HistoryLine::HistoryLine(int32 InitNr, int InitFromTower, int InitToTower, int InitDiscSize, int32 InitCheckedAgainst)
		: BStringItem(""), CheckedAgainst(InitCheckedAgainst), Nr(InitNr), FromTower(InitFromTower), ToTower(InitToTower), DiscSize(InitDiscSize) {
	char temp[100];
	sprintf(temp,"%li: Disc %i from %c to %c",Nr,DiscSize,(char)('A'+FromTower),(char)('A'+ToTower)); // create String for the list
	BStringItem::SetText(temp);
}

bool HistoryLine::IsUndoMove(HistoryLine *AnotherLine) {
	if (DiscSize!=(AnotherLine->DiscSize)) return false; // Another disc has been moved
	if (ToTower!=(AnotherLine->FromTower)) return false; // The Tower we're heading for was not the start tower of the other move
	if (FromTower!=(AnotherLine->ToTower)) return false; // The Tower come from was not the target tower of the other move
	return true; // Well looks like we really undid the other move ;-(
	}


void HistoryLine::MarkAsUndoMove() {
	char temp[100];
	sprintf(temp,"(%li: Disc %i from %c to %c)",Nr,DiscSize,(char)('A'+FromTower),(char)('A'+ToTower)); // create String for the list
	BStringItem::SetText(temp);
}		

//-----------------------------------------------------------------------------------------------------------------------------------------


// The constructor adds the counter (a BStringView) and the list with a scrollbar to the window and 
// show it. Of course, the data is reset ;-)
HistoryWindow::HistoryWindow(BRect frame, const char* name)
			: BWindow(frame, name,B_TITLED_WINDOW,B_NOT_CLOSABLE) {

	// Create and add Move Counter
	BRect r(10,10,Bounds().Width()-10,20);
	MoveCounter=new BStringView(r,"movecounter","Not initialized");
	AddChild(MoveCounter);

	// Create and add Info Line
	r.Set(10,25,Bounds().Width()-10,35);
	InfoLine=new BStringView(r,"InfoLine","Not initialized");
	AddChild(InfoLine);
	
	// Create the ListView for the move list
	r=Bounds();	// Take the whole window
	r.InsetBy(10,10); // Inset by 10 on all sides
	r.top=45; // set top to 25, because there's the MoveCounter
	r.right-=B_V_SCROLL_BAR_WIDTH; // leave some space for the Scrollbar!
	
	// Now create the ListView, but don't add it
	Moves=new BListView(r,"history", B_SINGLE_SELECTION_LIST, B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_NAVIGABLE | B_FRAME_EVENTS);

	// Instead, create a BScrollView, which will 'wrap around' the BListView
	// That BScrollView will create and add a Scrollbar for us and connect the scrollbar to the ListView
	// Add that ScrollView, which adds the rest in place.
	AddChild(new BScrollView("scroll_history", Moves, B_FOLLOW_ALL_SIDES, B_FRAME_EVENTS, false, true));

	// Reset everything
	Reset();
	
	// And show the beauty to the public
	Show();
}

// ShowMoveCounter converts the Counter (an integer) to a string and puts it on screen
void HistoryWindow::ShowMoveCounter() {
	char temp[20];
	sprintf(temp,"Moves done: %ld",MoveCount); // format and 'print' the value into the string
	//	printf("String: %s\n",temp); // Debug
	Lock(); // Lock the Window. This seems to be necessary before changing the Text (AFAIK)
	MoveCounter->SetText(temp); // change the text of the existing view
	Unlock(); // well.... unlock
}

// Add a Move to the History
// Note that the lines are added 'at the top' of the list, that means at index 0.
// This means, that the first move will have the highest index.
void HistoryWindow::AddMove(int FromTower, int ToTower, int DiscSize) {

//	printf("LastEffectiveMove=%li (before)\n",	LastEffectiveMove);// Debug

	int32 ThisMoveNr=Moves->CountItems()+1; // Nr. of this Move in the History List (just the number, not the index!)
	HistoryLine *HL=new HistoryLine(ThisMoveNr, FromTower, ToTower, DiscSize, LastEffectiveMove);
	if (LastEffectiveMove>=0) {
		HistoryLine* LEM_HL;
		int32 LEM_index=(Moves->CountItems())-LastEffectiveMove; // Count backwards, as Move Nr. 0 will have the higest index.
		LEM_HL=(HistoryLine*) Moves->ItemAt(LEM_index);
		if (HL->IsUndoMove(LEM_HL))	{ // Did we undo the last effective move?
			HL->MarkAsUndoMove();     // Oh yes.... then this is an undo move
			LEM_HL->MarkAsUndoMove(); // and the other one has just been undone
			Lock();
			Moves->InvalidateItem(LEM_index); // Take care that the old move needs to be redrawn.
			Unlock();
			LastEffectiveMove=LEM_HL->CheckedAgainst; // Trace for the last effective move before that
			}
		else // Move is not an Undo Move
			LastEffectiveMove=ThisMoveNr;
		}
	else // There were no effective moves up to now
		LastEffectiveMove=ThisMoveNr;
//	printf("LastEffectiveMove=%li (after)\n",	LastEffectiveMove); // Debug

	Lock();
	Moves->AddItem(HL,0); // Add string to list
//	Moves->AddItem(HL); // Add string to list (at the end)
	Unlock();
	MoveCount++;
	ShowMoveCounter(); // Increase and Show the counter
}

void HistoryWindow::Reset() {
	char temp[40];
	MoveCount=0;
	LastEffectiveMove=-1;
	int MinMoves = (1 << NumberOfDiscs)-1; // calc minimum (optimal) Number of Moves
	sprintf(temp,"%i Discs; Min. Moves: %i",NumberOfDiscs,MinMoves);
	ShowMoveCounter();

	Lock();
	InfoLine->SetText(temp);
	Moves->MakeEmpty(); // again, when you call MakeEmpty, lock the window
						// for details, see DiscTowerView::ResetDiscs()
	Unlock();
}

