//
// Towers of Hanoi
//
// Copyright 1998-1999
//
// Written by: Marco Zinn
//
// If you're about to read this source code, read Tower.h before, please....

#include <KernelKit.h>
#include <stdio.h>
#include "Towers.h" // defines the TowerApp class
#include "TowerWindow.h" // defines the Disc, DiscTowerView and TowerWindow classes
#include "ControlWindow.h" // defines the ControlWindow class
#include "HistoryWindow.h" // defines the HistoryWindow class

//
// Globale Vars:
//
int		NumberOfDiscs;		// Number of Discs currently in play (=Height of Tower when starting)
Disc	*Discs[MAX_DISCS+1];	// Array of Pointers to the (10) Disc Objects
Disc	*DummyDisc;			// Pointer to a special Dummy-Discs(tm). This is the space-holder on 'top' of the towers
// DummyDisc Remark: The towers that are visible on screen are Views (DiscTowerView, in fact. 
//                   see TowerWindow.h). To be precise, they are derived from BListView.
//                   Sadly for me, BListViews display a list top-down (if you don't want to 
//                   overwrite the draw()-part). In order to create a 'space' over the towers,
//                   so they all stand on a common level ground or base, I have to put 'empty'
//                   entries on top of the towers. These is the DummyDisc(tm). It's just invisble
//                   If you want to make it visible, see and change Disc::Draw()
//                   NOTE: As any BListItem (a Disc is derived from that) can occur in multiple
//                   Lists (better: BListViews) at the same time, we only need ONE DummyDisc(tm).
sem_id	SolvingSem;			// Semaphore for the solving thread
// Note to that: The auto-solving is done in a thread (named 'solver'), which is spawned, when
//  the function is started. The thread acquires this semaphore. On the other hand, the application
//  acquires this semaphore, when it is told to reset the tower. So, we are sure, that the tower
//  isn't reset while the solver runs. On the other hand, we are sure, that the solver doesn't get
//  started while the app still is resetting the towers. Both would up in a mess!

//
// Implementation of TowerApp class 
//
// This one is derived from BApplication. It will be our application.
// Note, that you won't see that class on screen. What you'll see are just the window
//

// The constructor does all the stuff that happens, when you start the app.
TowerApp::TowerApp()
			: BApplication("application/x-vnd.MarcoZ-Towers") { 
		// Note: If you're using a ressource file (as I do here), make sure,
		// the application signature here is the same as the one in the ressource file!

	NumberOfDiscs=4; // Height of the Start-Tower; is equal to the total number of discs in play
	
	// Create the Discs:
	//  At starttime, all (10) discs are created. How many get shown, is not relevant here.
	// 	By doing this, I do not have to create and destroy the discs every time the game
	//       will be reset by the user.
	rgb_color discColor; // Prepare the color the Discs will get:
	discColor.green=0;   // green, blue and alpha are the same for all discs (red will change)
	discColor.blue=0;
	discColor.alpha=128;

	// Loop and create all (10) discs now:
	for (int DiscCount=1;DiscCount<=MAX_DISCS;DiscCount++) { // loop from 1 to 10! (not from 0 to 9)
		discColor.red=(255-DiscCount*20); // red value changes. Run the app, you'll see it!
		Discs[DiscCount]= new Disc(DiscCount, discColor); // size, color
		}	

	// Create a single Dummy-Discs. They have color, but are NOT drawn. See disc::draw()
//	discColor.red=230;discColor.green=230;discColor.blue=230; // color was for debugging, of course
	discColor.red=0;discColor.green=0;discColor.blue=0; // color was for debugging, of course
	Discs[0] = new Disc (0,discColor); // 'size' of the Disc is 1. That's in case, you want to draw it.
	DummyDisc=Discs[0]; // DummyDisc is just an alias for Disc[0]. The two pointers just point to the same adress
	
	// We'll set up the windows now....
	// For a deeper explanation of the windows, see the appropriate header file!
	
	BRect windowRect; // That's the BRect we will (re-)use for the windows.

	// History Window for showing the moves made
	windowRect.Set(640,40,795,340);	
	History = new HistoryWindow(windowRect,"History");

	// Tower Window for showing the towers (and the menu).
	windowRect.Set(0,0,610,170); // Another way to place the window: This line is the size
	windowRect.OffsetTo(10,40);  // and that line is the position of the window. Easier to play around with ;-)
	Towers = new TowerWindow(windowRect,"Towers",History);
	
	// Control Window for auto-solving
	//  Supply the Pointers to the three Towers for Auto-Solving
	windowRect.Set(20,280,200,400);	
	Controls = new ControlWindow(windowRect,"Controls",Towers->LeftTowerView, Towers->RightTowerView, Towers->CenterTowerView);
	
	Towers->Activate(); // Bring this in front
}

// ResetGame resets the game (indeed!):
// it might set a new number of discs before doing that
// then it stacks up Tower A and clears the move history.
void TowerApp::ResetGame(int NewNumberOfDiscs) {
	Controls->Reset(); // Set Stop flag. If solver thread runs, this tells it to quit
	printf("Trying to reset game...\n"); // Debug
	if (acquire_sem(SolvingSem) != B_NO_ERROR) return; // Wait for the solver thread to exit
	if (NewNumberOfDiscs>MAX_DISCS) NewNumberOfDiscs=MAX_DISCS; // Error-Check.
	NumberOfDiscs=NewNumberOfDiscs;
	printf("\nApp::Reset: NumberOfDiscs=%i\n",NumberOfDiscs); // Debug
	// Stop Auto-Solving (in case it's running) and make sure the labels of the menu and
	//  the button is right.
	History->Reset(); // Sets the Counter to 0 and clears the history list
	Towers->Reset(); // resets the Towers (A will be full, B and C empty)
	printf("App::Reset: done\n"); // Debug
	release_sem(SolvingSem);
	}

// The destructor destructs the discs (Discs Array and DummyDisc)
// Note, that you shouldn't destroy the windows. AFAIK they are destroyed by some
// other magical mechanism!
TowerApp::~TowerApp() {
	// Destroy the dummyDisc
	delete DummyDisc; // Discs[0] is also invalid now!
	
	// And destroy the other, real discs
	for (int DiscCount=1;DiscCount<=MAX_DISCS;DiscCount++) // loop from 1 to 10
		delete Discs[DiscCount]; // Destroys the array...

}

// AboutRequested is called when the About window is asked for. It brings up...
// ... the About-Box (believe it or not!)
//  see TowerWindow.c for how the menu works.
void TowerApp::AboutRequested()
{
	(new BAlert("About...",
		"Towers of Benoi V1.7\n\n"
		"The first thing I can publish without getting completely red...\n"
		"Marco Zinn, 1998-99 (marco.zinn@gmx.de)",
		"Very Cool!"))->Go();
	// .... well, I copied that from somewhere, but it works and it's short...
}


// The application only gets the messages concerning the application itself (ie. the game)
//  In detail, these are RESET_GAME and SOLVE_GAME requests from the menu.
void TowerApp::MessageReceived(BMessage *message){
	int i;
	switch(message->what) {
		case RESET_GAME: // application got a request to reset the game
			if (Towers->LeftTowerView->Window()==NULL) break;
			i=message->FindInt32("Discs"); // retrieve number of discs to reset the game to
			// printf("Reset to %i discs.\n",i); // Debug again
			if ((i>1) && (i<=10)) ResetGame(i);	 // Error-Check and reset.
			if (i==0) ResetGame(); // will reset without changing the number of discs.
			break;
		case START_SOLVING: // application got a request to solve the game
			printf("Got start message\n");
			Towers->PostMessage(message); // send the same message to the Tower Window (updates menu item)
			Controls->StartSolving();
			break;
		case STOP_SOLVING: // application got a request to solve the game
			printf("Got stop message\n");
			Towers->PostMessage(message); // send the same message to the Tower Window (updates menu item)
			Controls->Reset();
			break;
		case SHOW_DOCS: // request to show the docs		
			{
			status_t DummyStatus;
			app_info MyInfo;
			entry_ref MyRef;
			DummyStatus=GetAppInfo(&MyInfo);
			BEntry MyEntry(&(MyInfo.ref));
			BDirectory MyDir;
			MyEntry.GetParent(&MyDir);
			
//			BPath MyPath(&MyEntry);
//			char AppName[B_FILE_NAME_LENGTH];
//			MyEntry.GetName(AppName);
//			printf("Documentation requested. AppName = %s\n",AppName);
//			printf("Documentation requested. Path = %s\n",MyPath.Path());
			
			BEntry rules(&MyDir,"Documentation/towers_docu.htm");
			if (rules.GetRef(&MyRef)==B_OK)
				be_roster->Launch(&MyRef);
			}
			break;
		default: // Any other messages are passed to BApplication. DO NOT FORGET THIS!
			BApplication::MessageReceived(message);
			break;
	}
}

//
// main
//

// Uh. oh. Well, that's the main programm. guess, you've seen something like that before...
int main(void) {
	TowerApp theApp;
	theApp.Run();
}
 